from __future__ import annotations

import collections.abc
import json
import os
from dataclasses import dataclass

from enum import Enum, auto
from typing import TYPE_CHECKING

from .parsers import *


if TYPE_CHECKING:
    from typing import Any, Dict, Iterator, List, Literal, Optional, Union


__VERSION__ = '0.4.0'
__DATE__ = '2021-11-08'
__all__ = [
    'EnvProps',
    'EnvPropsError',
    'EnvPropsParseError',
    'EnvPropsValueNotDefined',
    'Parsers',
    'Property',
    'RedactedProperty',
    'ValueState',
]


try:
    import yaml
    yaml_supported = True
except ModuleNotFoundError:
    yaml_supported = False


class ValueState(Enum):
    NOTPROVIDED = auto()
    HIDDEN = auto()
    NONE = auto()


class EnvPropsError(Exception):
    pass


class EnvPropsParseError(EnvPropsError):
    def __init__(self, property: Property, value: Any, error: Exception) -> None:
        self.property = property
        self.value = value
        self.error = error
        super().__init__(f'value could not be parsed:\nproperty: {self.property}\nvalue: {self.value}\nerror: {type(self.error).__name__}: {self.error}')


class EnvPropsValueNotDefined(EnvPropsError):
    def __init__(self, property: Property) -> None:
        self.property = property
        super().__init__(f'value is required but not specified by default or environment\n{self.property}')


@dataclass
class Property:
    name: str
    type: str
    required: bool
    sensitive: bool
    default: Any
    environment: Any

    def value(self) -> Any:
        raw_value = self.default if self.environment is ValueState.NOTPROVIDED else self.environment
        if self.required is True and raw_value is ValueState.NOTPROVIDED:
            raise EnvPropsValueNotDefined(self)
        else:
            if raw_value is ValueState.NOTPROVIDED or raw_value is None:
                return None
            else:
                try:
                    return self.parse(raw_value)
                except EnvPropsError as e:
                    raise e
                except Exception as e:
                    raise EnvPropsParseError(self, raw_value, e)

    def parse(self, value: str) -> Any:
        try:
            return Parsers[self.type.upper()](value)
        except KeyError:
            raise EnvPropsError(f'type={self.type} not supported [available values: {",".join(Parsers.keys())}]')

    def redact(self) -> Union[Property, RedactedProperty]:
        if self.sensitive is True:
            return RedactedProperty(
                name=self.name,
                type=self.type,
                required=self.required,
                sensitive=self.sensitive,
                default=ValueState.NOTPROVIDED if self.default is ValueState.NOTPROVIDED else ValueState.HIDDEN,
                environment=ValueState.NOTPROVIDED if self.environment is ValueState.NOTPROVIDED else ValueState.HIDDEN,
            )
        else:
            return self

    @staticmethod
    def new(name: str, definition: dict) -> Property:
        if 'type' not in definition:
            raise KeyError(f'type is required for {name}')

        return Property(
            name=name,
            type=definition['type'],
            required=definition.get('required', True),
            sensitive=definition.get('sensitive', False),
            default=definition.get('default', ValueState.NOTPROVIDED),
            environment=os.environ.get(name, ValueState.NOTPROVIDED),
        )


@dataclass
class RedactedProperty(Property):
    def value(self) -> Literal[ValueState.HIDDEN]:
        return ValueState.HIDDEN


class EnvProps(collections.abc.Mapping):
    def __init__(self, filename: str, yaml_loader: str = 'SafeLoader') -> None:
        super().__init__()
        self.filename = filename
        self.yaml_loader = yaml_loader
        self.properties = self.parse()

    def __iter__(self) -> Iterator[str]:
        return iter(self.properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.properties[key].value()

    def __len__(self) -> int:
        return len(self.properties)

    def init_app(self, app) -> None:
        """
            helper method to add properties to Flask or Quart
        """
        for name, prop in self.properties.items():
            redacted_prop = prop.redact()
            redacted_value = redacted_prop.value()
            redacted_value = ValueState.NONE if redacted_value is None else redacted_value
            app.logger.info(f'{name}: {redacted_value}')
        app.config.from_mapping(self)

    def read(self) -> Dict[str, Dict[str, Any]]:
        with open(self.filename, 'r') as fh:
            data = fh.read()

        if self.filename.endswith('.yaml'):
            if yaml_supported is False:
                raise EnvPropsError('pyyaml is not available; please run "pip install pyyaml"')
            yaml_loader = getattr(yaml, self.yaml_loader) if self.yaml_loader else None
            return yaml.load(data, Loader=yaml_loader)
        elif self.filename.endswith('.json'):
            return json.loads(data)
        else:
            raise EnvPropsError('supported file types are json and yaml')

    def parse(self) -> Dict[str, Property]:
        data = self.read()
        property_dict = dict()
        if 'properties' not in data:
            raise EnvPropsError('the "properties" key is missing from {self.filename}')
        for name, definition in data['properties'].items():
            property_dict[name] = Property.new(name, definition)
        return property_dict
