from __future__ import annotations

import json
from base64 import b64decode
from datetime import datetime
from distutils.util import strtobool
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING
from uuid import UUID


if TYPE_CHECKING:
    from typing import Any, Dict, Iterator, List, Literal, Mapping, Optional, Union


__all__ = [
    'Parsers'
]


try:
    import dateparser
    date_supported = True
except ModuleNotFoundError:
    date_supported = False


def _parse_string(value: str) -> str:
    value = str(value)
    if value.startswith('base64:'):
        return b64decode(value[7:]).decode('utf-8')
    else:
        return value


def _parse_bytes(value: str) -> bytes:
    value = str(value)
    if value.startswith('base64:'):
        return b64decode(value[7:])
    else:
        return value.encode('utf-8')


def _parse_int(value: str) -> int:
    return int(value)


def _parse_float(value: str) -> float:
    return float(value)


def _parse_bool(value: str) -> bool:
    if type(value) is str:
        return bool(strtobool(value))
    else:
        return bool(value)


def _parse_uuid(value: str) -> UUID:
    return UUID(value)


def _parse_json(value: str) -> Mapping:
    return json.loads(value)


def _parse_date(value: str) -> Optional[datetime]:
    if date_supported is False:
        raise ValueError('dateparser is not available; please run "pip install dateparser"')
    return dateparser.parse(value)


class Parsers(Enum):
    BOOL = partial(_parse_bool)
    BYTES = partial(_parse_bytes)
    DATE = partial(_parse_date)
    FLOAT = partial(_parse_float)
    INT = partial(_parse_int)
    JSON = partial(_parse_json)
    STR = partial(_parse_string)
    UUID = partial(_parse_uuid)

    def __call__(self, *args, **kwargs) -> Any:
        return self.value(*args, **kwargs)

    @staticmethod
    def keys() -> List[str]:
        return [k.lower() for k in Parsers.__members__.keys()]
